@@ -10,7 +10,7 @@
  *
  * The rest of the file is licensed under the BSD license.  See LICENSE.
  *
- * $Id: assoc.c 595 2007-07-10 06:30:07Z plindner $
+ * $Id$
  */
 
 #include "memcached.h"
@@ -476,13 +476,11 @@ static bool expanding = false;
 static unsigned int expand_bucket = 0;
 
 void assoc_init(void) {
-    unsigned int hash_size = hashsize(hashpower) * sizeof(void*);
-    primary_hashtable = malloc(hash_size);
+    primary_hashtable = calloc(hashsize(hashpower), sizeof(void *));
     if (! primary_hashtable) {
         fprintf(stderr, "Failed to init hashtable.\n");
         exit(EXIT_FAILURE);
     }
-    memset(primary_hashtable, 0, hash_size);
 }
 
 item *assoc_find(const char *key, const size_t nkey) {
@@ -498,14 +496,18 @@ item *assoc_find(const char *key, const size_t nkey) {
         it = primary_hashtable[hv & hashmask(hashpower)];
     }
 
+    item *ret = NULL;
+    int depth = 0;
     while (it) {
-        if ((nkey == it->nkey) &&
-            (memcmp(key, ITEM_key(it), nkey) == 0)) {
-            return it;
+        if ((nkey == it->nkey) && (memcmp(key, ITEM_key(it), nkey) == 0)) {
+            ret = it;
+            break;
         }
         it = it->h_next;
+        ++depth;
     }
-    return 0;
+    MEMCACHED_ASSOC_FIND(key, depth);
+    return ret;
 }
 
 /* returns the address of the item pointer before the key.  if *item == 0,
@@ -597,6 +599,7 @@ int assoc_insert(item *it) {
         assoc_expand();
     }
 
+    MEMCACHED_ASSOC_INSERT(ITEM_key(it), hash_items);
     return 1;
 }
 
@@ -604,10 +607,15 @@ void assoc_delete(const char *key, const size_t nkey) {
     item **before = _hashitem_before(key, nkey);
 
     if (*before) {
-        item *nxt = (*before)->h_next;
+        item *nxt;
+        hash_items--;
+        /* The DTrace probe cannot be triggered as the last instruction
+         * due to possible tail-optimization by the compiler
+         */
+        MEMCACHED_ASSOC_DELETE(key, hash_items);
+        nxt = (*before)->h_next;
         (*before)->h_next = 0;   /* probably pointless, but whatever. */
         *before = nxt;
-        hash_items--;
         return;
     }
     /* Note:  we never actually get here.  the callers don't delete things
