@@ -80,7 +80,6 @@ static void settings_init(void);
 static void event_handler(const int fd, const short which, void *arg);
 static void conn_close(conn *c);
 static void conn_init(void);
-static void accept_new_conns(const bool do_accept);
 static bool update_event(conn *c, const int new_flags);
 static void complete_nread(conn *c);
 static void process_command(conn *c, char *command);
@@ -112,8 +111,6 @@ static struct event_base *main_base;
 #define TRANSMIT_SOFT_ERROR 2
 #define TRANSMIT_HARD_ERROR 3
 
-static int *buckets = 0; /* bucket->generation array for a managed instance */
-
 #define REALTIME_MAXDELTA 60*60*24*30
 /*
  * given time value that's either unix time or delta from current unix time, return
@@ -143,7 +140,9 @@ static rel_time_t realtime(const time_t exptime) {
 static void stats_init(void) {
     stats.curr_items = stats.total_items = stats.curr_conns = stats.total_conns = stats.conn_structs = 0;
     stats.get_cmds = stats.set_cmds = stats.get_hits = stats.get_misses = stats.evictions = 0;
-    stats.curr_bytes = stats.bytes_read = stats.bytes_written = 0;
+    stats.curr_bytes = stats.bytes_read = stats.bytes_written = stats.flush_cmds = 0;
+    stats.listen_disabled_num = 0;
+    stats.accepting_conns = 1; /* assuming we start in this state. */
 
     /* make the time we started always be 2 seconds before we really
        did, so time(0) - time.started is never zero.  if so, things
@@ -158,6 +157,7 @@ static void stats_reset(void) {
     stats.total_items = stats.total_conns = 0;
     stats.get_cmds = stats.set_cmds = stats.get_hits = stats.get_misses = stats.evictions = 0;
     stats.bytes_read = stats.bytes_written = 0;
+    stats.flush_cmds = stats.listen_disabled_num = 0;
     stats_prefix_clear();
     STATS_UNLOCK();
 }
@@ -165,7 +165,7 @@ static void stats_reset(void) {
 static void settings_init(void) {
     settings.access=0700;
     settings.port = 11211;
-    settings.udpport = 0;
+    settings.udpport = 11211;
     /* By default this string should be NULL for getaddrinfo() */
     settings.inter = NULL;
     settings.maxbytes = 64 * 1024 * 1024; /* default is 64MB */
@@ -174,7 +174,6 @@ static void settings_init(void) {
     settings.oldest_live = 0;
     settings.evict_to_free = 1;       /* push old items out of cache when memory runs out */
     settings.socketpath = NULL;       /* by default, not using a unix socket */
-    settings.managed = false;
     settings.factor = 1.25;
     settings.chunk_size = 48;         /* space for a modest key and value */
 #ifdef USE_THREADS
@@ -182,8 +181,11 @@ static void settings_init(void) {
 #else
     settings.num_threads = 1;
 #endif
+    settings.num_threads++;  /* N workers + 1 dispatcher */
     settings.prefix_delimiter = ':';
     settings.detail_enabled = 0;
+    settings.reqs_per_event = 20;
+    settings.backlog = 1024;
 }
 
 /* returns true if a deleted item's delete-locked-time is over, and it
@@ -365,8 +367,6 @@ conn *conn_new(const int sfd, const int init_state, const int event_flags,
     c->write_and_go = conn_read;
     c->write_and_free = 0;
     c->item = 0;
-    c->bucket = -1;
-    c->gen = 0;
 
     c->noreply = false;
 
@@ -1062,7 +1062,7 @@ static void process_stat(conn *c, token_t *tokens, const size_t ntokens) {
     command = tokens[COMMAND_TOKEN].value;
 
     if (ntokens == 2 && strcmp(command, "stats") == 0) {
-        char temp[1024];
+        char temp[2048];
         pid_t pid = getpid();
         char *pos = temp;
 
@@ -1087,6 +1087,7 @@ static void process_stat(conn *c, token_t *tokens, const size_t ntokens) {
         pos += sprintf(pos, "STAT curr_connections %u\r\n", stats.curr_conns - 1); /* ignore listening conn */
         pos += sprintf(pos, "STAT total_connections %u\r\n", stats.total_conns);
         pos += sprintf(pos, "STAT connection_structures %u\r\n", stats.conn_structs);
+        pos += sprintf(pos, "STAT cmd_flush %llu\r\n", stats.flush_cmds);
         pos += sprintf(pos, "STAT cmd_get %llu\r\n", stats.get_cmds);
         pos += sprintf(pos, "STAT cmd_set %llu\r\n", stats.set_cmds);
         pos += sprintf(pos, "STAT get_hits %llu\r\n", stats.get_hits);
@@ -1096,6 +1097,8 @@ static void process_stat(conn *c, token_t *tokens, const size_t ntokens) {
         pos += sprintf(pos, "STAT bytes_written %llu\r\n", stats.bytes_written);
         pos += sprintf(pos, "STAT limit_maxbytes %llu\r\n", (uint64_t) settings.maxbytes);
         pos += sprintf(pos, "STAT threads %u\r\n", settings.num_threads);
+        pos += sprintf(pos, "STAT accepting_conns %u\r\n", stats.accepting_conns);
+        pos += sprintf(pos, "STAT listen_disabled_num %llu\r\n", stats.listen_disabled_num);
         pos += sprintf(pos, "END");
         STATS_UNLOCK();
         out_string(c, temp);
@@ -1134,43 +1137,6 @@ static void process_stat(conn *c, token_t *tokens, const size_t ntokens) {
 #endif /* HAVE_STRUCT_MALLINFO */
 #endif /* HAVE_MALLOC_H */
 
-#if !defined(WIN32) || !defined(__APPLE__)
-    if (strcmp(subcommand, "maps") == 0) {
-        char *wbuf;
-        int wsize = 8192; /* should be enough */
-        int fd;
-        int res;
-
-        if ((wbuf = (char *)malloc(wsize)) == NULL) {
-            out_string(c, "SERVER_ERROR out of memory writing stats maps");
-            return;
-        }
-
-        fd = open("/proc/self/maps", O_RDONLY);
-        if (fd == -1) {
-            out_string(c, "SERVER_ERROR cannot open the maps file");
-            free(wbuf);
-            return;
-        }
-
-        res = read(fd, wbuf, wsize - 6);  /* 6 = END\r\n\0 */
-        if (res == wsize - 6) {
-            out_string(c, "SERVER_ERROR buffer overflow");
-            free(wbuf); close(fd);
-            return;
-        }
-        if (res == 0 || res == -1) {
-            out_string(c, "SERVER_ERROR can't read the maps file");
-            free(wbuf); close(fd);
-            return;
-        }
-        memcpy(wbuf + res, "END\r\n", 5);
-        write_and_free(c, wbuf, res + 5);
-        close(fd);
-        return;
-    }
-#endif
-
     if (strcmp(subcommand, "cachedump") == 0) {
 
         char *buf;
@@ -1239,19 +1205,6 @@ static inline void process_get_command(conn *c, token_t *tokens, size_t ntokens,
     int stats_get_misses = 0;
     assert(c != NULL);
 
-    if (settings.managed) {
-        int bucket = c->bucket;
-        if (bucket == -1) {
-            out_string(c, "CLIENT_ERROR no BG data in managed mode");
-            return;
-        }
-        c->bucket = -1;
-        if (buckets[bucket] != c->gen) {
-            out_string(c, "ERROR_NOT_OWNER");
-            return;
-        }
-    }
-
     do {
         while(key_token->length != 0) {
 
@@ -1453,19 +1406,6 @@ static void process_update_command(conn *c, token_t *tokens, const size_t ntoken
         stats_prefix_record_set(key);
     }
 
-    if (settings.managed) {
-        int bucket = c->bucket;
-        if (bucket == -1) {
-            out_string(c, "CLIENT_ERROR no BG data in managed mode");
-            return;
-        }
-        c->bucket = -1;
-        if (buckets[bucket] != c->gen) {
-            out_string(c, "ERROR_NOT_OWNER");
-            return;
-        }
-    }
-
     it = item_alloc(key, nkey, flags, realtime(exptime), vlen+2);
 
     if (it == 0) {
@@ -1518,19 +1458,6 @@ static void process_arithmetic_command(conn *c, token_t *tokens, const size_t nt
     key = tokens[KEY_TOKEN].value;
     nkey = tokens[KEY_TOKEN].length;
 
-    if (settings.managed) {
-        int bucket = c->bucket;
-        if (bucket == -1) {
-            out_string(c, "CLIENT_ERROR no BG data in managed mode");
-            return;
-        }
-        c->bucket = -1;
-        if (buckets[bucket] != c->gen) {
-            out_string(c, "ERROR_NOT_OWNER");
-            return;
-        }
-    }
-
     delta = strtoll(tokens[2].value, NULL, 10);
 
     if(errno == ERANGE) {
@@ -1561,7 +1488,7 @@ static void process_arithmetic_command(conn *c, token_t *tokens, const size_t nt
  */
 char *do_add_delta(conn *c, item *it, const bool incr, const int64_t delta, char *buf) {
     char *ptr;
-    int64_t value;
+    uint64_t value;
     int res;
 
     ptr = ITEM_data(it);
@@ -1594,6 +1521,10 @@ char *do_add_delta(conn *c, item *it, const bool incr, const int64_t delta, char
         do_item_replace(it, new_it);
         do_item_remove(new_it);       /* release our reference */
     } else { /* replace in-place */
+        /* When changing the value without replacing the item, we
+           need to update the CAS on the existing item. */
+        it->cas_id = get_cas_id();
+
         memcpy(ITEM_data(it), buf, res);
         memset(ITEM_data(it) + res, ' ', it->nbytes - res - 2);
     }
@@ -1611,19 +1542,6 @@ static void process_delete_command(conn *c, token_t *tokens, const size_t ntoken
 
     set_noreply_maybe(c, tokens, ntokens);
 
-    if (settings.managed) {
-        int bucket = c->bucket;
-        if (bucket == -1) {
-            out_string(c, "CLIENT_ERROR no BG data in managed mode");
-            return;
-        }
-        c->bucket = -1;
-        if (buckets[bucket] != c->gen) {
-            out_string(c, "ERROR_NOT_OWNER");
-            return;
-        }
-    }
-
     key = tokens[KEY_TOKEN].value;
     nkey = tokens[KEY_TOKEN].length;
 
@@ -1766,67 +1684,6 @@ static void process_command(conn *c, char *command) {
 
         process_delete_command(c, tokens, ntokens);
 
-    } else if (ntokens == 3 && strcmp(tokens[COMMAND_TOKEN].value, "own") == 0) {
-        unsigned int bucket, gen;
-        if (!settings.managed) {
-            out_string(c, "CLIENT_ERROR not a managed instance");
-            return;
-        }
-
-        if (sscanf(tokens[1].value, "%u:%u", &bucket,&gen) == 2) {
-            if ((bucket < 0) || (bucket >= MAX_BUCKETS)) {
-                out_string(c, "CLIENT_ERROR bucket number out of range");
-                return;
-            }
-            buckets[bucket] = gen;
-            out_string(c, "OWNED");
-            return;
-        } else {
-            out_string(c, "CLIENT_ERROR bad format");
-            return;
-        }
-
-    } else if (ntokens == 3 && (strcmp(tokens[COMMAND_TOKEN].value, "disown")) == 0) {
-
-        int bucket;
-        if (!settings.managed) {
-            out_string(c, "CLIENT_ERROR not a managed instance");
-            return;
-        }
-        if (sscanf(tokens[1].value, "%u", &bucket) == 1) {
-            if ((bucket < 0) || (bucket >= MAX_BUCKETS)) {
-                out_string(c, "CLIENT_ERROR bucket number out of range");
-                return;
-            }
-            buckets[bucket] = 0;
-            out_string(c, "DISOWNED");
-            return;
-        } else {
-            out_string(c, "CLIENT_ERROR bad format");
-            return;
-        }
-
-    } else if (ntokens == 3 && (strcmp(tokens[COMMAND_TOKEN].value, "bg")) == 0) {
-        int bucket, gen;
-        if (!settings.managed) {
-            out_string(c, "CLIENT_ERROR not a managed instance");
-            return;
-        }
-        if (sscanf(tokens[1].value, "%u:%u", &bucket, &gen) == 2) {
-            /* we never write anything back, even if input's wrong */
-            if ((bucket < 0) || (bucket >= MAX_BUCKETS) || (gen <= 0)) {
-                /* do nothing, bad input */
-            } else {
-                c->bucket = bucket;
-                c->gen = gen;
-            }
-            conn_set_state(c, conn_read);
-            return;
-        } else {
-            out_string(c, "CLIENT_ERROR bad format");
-            return;
-        }
-
     } else if (ntokens >= 2 && (strcmp(tokens[COMMAND_TOKEN].value, "stats") == 0)) {
 
         process_stat(c, tokens, ntokens);
@@ -1837,6 +1694,10 @@ static void process_command(conn *c, char *command) {
 
         set_noreply_maybe(c, tokens, ntokens);
 
+        STATS_LOCK();
+        stats.flush_cmds++;
+        STATS_UNLOCK();
+
         if(ntokens == (c->noreply ? 3 : 2)) {
             settings.oldest_live = current_time - 1;
             item_flush_expired();
@@ -2069,16 +1930,13 @@ static bool update_event(conn *c, const int new_flags) {
 /*
  * Sets whether we are listening for new connections or not.
  */
-void accept_new_conns(const bool do_accept) {
+void do_accept_new_conns(const bool do_accept) {
     conn *next;
 
-    if (! is_listen_thread())
-        return;
-
     for (next = listen_conn; next; next = next->next) {
         if (do_accept) {
             update_event(next, EV_READ | EV_PERSIST);
-            if (listen(next->sfd, 1024) != 0) {
+            if (listen(next->sfd, settings.backlog) != 0) {
                 perror("listen");
             }
         }
@@ -2088,7 +1946,18 @@ void accept_new_conns(const bool do_accept) {
                 perror("listen");
             }
         }
-  }
+    }
+
+    if (do_accept) {
+        STATS_LOCK();
+        stats.accepting_conns = 1;
+        STATS_UNLOCK();
+    } else {
+        STATS_LOCK();
+        stats.accepting_conns = 0;
+        stats.listen_disabled_num++;
+        STATS_UNLOCK();
+    }
 }
 
 
@@ -2164,6 +2033,7 @@ static void drive_machine(conn *c) {
     int sfd, flags = 1;
     socklen_t addrlen;
     struct sockaddr_storage addr;
+    int nreqs = settings.reqs_per_event;
     int res;
 
     assert(c != NULL);
@@ -2202,7 +2072,9 @@ static void drive_machine(conn *c) {
             if (try_read_command(c) != 0) {
                 continue;
             }
-            if ((c->udp ? try_read_udp(c) : try_read_network(c)) != 0) {
+            /* Only process nreqs at a time to avoid starving other
+               connections */
+            if (--nreqs && (c->udp ? try_read_udp(c) : try_read_network(c)) != 0) {
                 continue;
             }
             /* we have no command line and no data to read from network */
@@ -2409,7 +2281,6 @@ static int new_socket(struct addrinfo *ai) {
     int flags;
 
     if ((sfd = socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol)) == -1) {
-        perror("socket()");
         return -1;
     }
 
@@ -2474,43 +2345,61 @@ static int server_socket(const int port, const bool is_udp) {
      * that otherwise mess things up.
      */
     memset(&hints, 0, sizeof (hints));
-    hints.ai_flags = AI_PASSIVE|AI_ADDRCONFIG;
+    hints.ai_flags  = AI_PASSIVE;
+    hints.ai_family = AF_UNSPEC;
     if (is_udp)
     {
-        hints.ai_protocol = IPPROTO_UDP;
         hints.ai_socktype = SOCK_DGRAM;
-        hints.ai_family = AF_INET; /* This left here because of issues with OSX 10.5 */
     } else {
-        hints.ai_family = AF_UNSPEC;
-        hints.ai_protocol = IPPROTO_TCP;
         hints.ai_socktype = SOCK_STREAM;
     }
 
     snprintf(port_buf, NI_MAXSERV, "%d", port);
     error= getaddrinfo(settings.inter, port_buf, &hints, &ai);
     if (error != 0) {
-      if (error != EAI_SYSTEM)
-        fprintf(stderr, "getaddrinfo(): %s\n", gai_strerror(error));
-      else
-        perror("getaddrinfo()");
+        if (error != EAI_SYSTEM)
+            fprintf(stderr, "getaddrinfo(): %s\n", gai_strerror(error));
+        else
+            perror("getaddrinfo()");
 
-      return 1;
+        return 1;
     }
 
     for (next= ai; next; next= next->ai_next) {
         conn *listen_conn_add;
         if ((sfd = new_socket(next)) == -1) {
-            freeaddrinfo(ai);
-            return 1;
+            /* getaddrinfo can return "junk" addresses,
+             * we make sure at least one works before erroring.
+             */
+            continue;
         }
 
+#ifdef IPV6_V6ONLY
+        if (next->ai_family == AF_INET6) {
+            error = setsockopt(sfd, IPPROTO_IPV6, IPV6_V6ONLY, (char *) &flags, sizeof(flags));
+            if (error != 0) {
+                perror("setsockopt");
+                close(sfd);
+                continue;
+            }
+        }
+#endif
+
         setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, (void *)&flags, sizeof(flags));
         if (is_udp) {
             maximize_sndbuf(sfd);
         } else {
-            setsockopt(sfd, SOL_SOCKET, SO_KEEPALIVE, (void *)&flags, sizeof(flags));
-            setsockopt(sfd, SOL_SOCKET, SO_LINGER, (void *)&ling, sizeof(ling));
-            setsockopt(sfd, IPPROTO_TCP, TCP_NODELAY, (void *)&flags, sizeof(flags));
+            error = setsockopt(sfd, SOL_SOCKET, SO_KEEPALIVE, (void *)&flags, sizeof(flags));
+            if (error != 0)
+                perror("setsockopt");
+
+            error = setsockopt(sfd, SOL_SOCKET, SO_LINGER, (void *)&ling, sizeof(ling));
+            if (error != 0)
+                perror("setsockopt");
+
+            error = setsockopt(sfd, IPPROTO_TCP, TCP_NODELAY, (void *)&flags, sizeof(flags));
+            if (error != 0)
+                perror("setsockopt");
         }
 
         if (bind(sfd, next->ai_addr, next->ai_addrlen) == -1) {
@@ -2523,34 +2412,34 @@ static int server_socket(const int port, const bool is_udp) {
             close(sfd);
             continue;
         } else {
-          success++;
-          if (!is_udp && listen(sfd, 1024) == -1) {
-              perror("listen()");
-              close(sfd);
-              freeaddrinfo(ai);
-              return 1;
+            success++;
+            if (!is_udp && listen(sfd, settings.backlog) == -1) {
+                perror("listen()");
+                close(sfd);
+                freeaddrinfo(ai);
+                return 1;
+            }
+        }
+
+        if (is_udp)
+        {
+          int c;
+
+          for (c = 1; c < settings.num_threads; c++) {
+              /* this is guaranteed to hit all threads because we round-robin */
+              dispatch_conn_new(sfd, conn_read, EV_READ | EV_PERSIST,
+                                UDP_READ_BUFFER_SIZE, is_udp);
+          }
+        } else {
+          if (!(listen_conn_add = conn_new(sfd, conn_listening,
+                                           EV_READ | EV_PERSIST, 1, false, main_base))) {
+              fprintf(stderr, "failed to create listening connection\n");
+              exit(EXIT_FAILURE);
           }
-      }
 
-      if (is_udp)
-      {
-        int c;
-
-        for (c = 0; c < settings.num_threads; c++) {
-            /* this is guaranteed to hit all threads because we round-robin */
-            dispatch_conn_new(sfd, conn_read, EV_READ | EV_PERSIST,
-                              UDP_READ_BUFFER_SIZE, 1);
+          listen_conn_add->next = listen_conn;
+          listen_conn = listen_conn_add;
         }
-      } else {
-        if (!(listen_conn_add = conn_new(sfd, conn_listening,
-                                         EV_READ | EV_PERSIST, 1, false, main_base))) {
-            fprintf(stderr, "failed to create listening connection\n");
-            exit(EXIT_FAILURE);
-        }
-
-        listen_conn_add->next = listen_conn;
-        listen_conn = listen_conn_add;
-      }
     }
 
     freeaddrinfo(ai);
@@ -2621,7 +2510,7 @@ static int server_socket_unix(const char *path, int access_mask) {
         return 1;
     }
     umask(old_umask);
-    if (listen(sfd, 1024) == -1) {
+    if (listen(sfd, settings.backlog) == -1) {
         perror("listen()");
         close(sfd);
         return 1;
@@ -2714,7 +2603,7 @@ void do_run_deferred_deletes(void)
 static void usage(void) {
     printf(PACKAGE " " VERSION "\n");
     printf("-p <num>      TCP port number to listen on (default: 11211)\n"
-           "-U <num>      UDP port number to listen on (default: 0, off)\n"
+           "-U <num>      UDP port number to listen on (default: 11211, 0 is off)\n"
            "-s <file>     unix socket path to listen on (disables network support)\n"
            "-a <mask>     access mask for unix socket, in octal (default 0700)\n"
            "-l <ip_addr>  interface to listen on, default is INDRR_ANY\n"
@@ -2734,7 +2623,6 @@ static void usage(void) {
            "-vv           very verbose (also print client commands/reponses)\n"
            "-h            print this help and exit\n"
            "-i            print memcached and libevent license\n"
-           "-b            run a managed instanced (mnemonic: buckets)\n"
            "-P <file>     save PID in <file>, only used with -d option\n"
            "-f <factor>   chunk size growth factor, default 1.25\n"
            "-n <bytes>    minimum space allocated for key+value+flags, default 48\n"
@@ -2751,6 +2639,10 @@ static void usage(void) {
 #ifdef USE_THREADS
     printf("-t <num>      number of threads to use, default 4\n");
 #endif
+    printf("-R            Maximum number of requests per event\n"
+           "              limits the number of requests process for a given con nection\n"
+           "              to prevent starvation.  default 20\n");
+    printf("-b            Set the backlog queue limit (default 1024)\n");
     return;
 }
 
@@ -2928,7 +2820,7 @@ int main (int argc, char **argv) {
     setbuf(stderr, NULL);
 
     /* process arguments */
-    while ((c = getopt(argc, argv, "a:bp:s:U:m:Mc:khirvdl:u:P:f:s:n:t:D:L")) != -1) {
+    while ((c = getopt(argc, argv, "a:p:s:U:m:Mc:khirvdl:u:P:f:s:n:t:D:LR:b:")) != -1) {
         switch (c) {
         case 'a':
             /* access for unix domain socket, as octal mask (like chmod)*/
@@ -2938,9 +2830,6 @@ int main (int argc, char **argv) {
         case 'U':
             settings.udpport = atoi(optarg);
             break;
-        case 'b':
-            settings.managed = true;
-            break;
         case 'p':
             settings.port = atoi(optarg);
             break;
@@ -2977,6 +2866,13 @@ int main (int argc, char **argv) {
         case 'r':
             maxcore = 1;
             break;
+        case 'R':
+            settings.reqs_per_event = atoi(optarg);
+            if (settings.reqs_per_event == 0) {
+                fprintf(stderr, "Number of requests per event must be greater than 0\n");
+                return 1;
+            }
+            break;
         case 'u':
             username = optarg;
             break;
@@ -2998,8 +2894,8 @@ int main (int argc, char **argv) {
             }
             break;
         case 't':
-            settings.num_threads = atoi(optarg);
-            if (settings.num_threads == 0) {
+            settings.num_threads = atoi(optarg) + 1; /* Extra dispatch thread */
+            if (settings.num_threads < 2) {
                 fprintf(stderr, "Number of threads must be greater than 0\n");
                 return 1;
             }
@@ -3019,6 +2915,9 @@ int main (int argc, char **argv) {
             }
             break;
 #endif
+        case 'b' :
+            settings.backlog = atoi(optarg);
+            break;
         default:
             fprintf(stderr, "Illegal argument \"%c\"\n", c);
             return 1;
@@ -3123,16 +3022,6 @@ int main (int argc, char **argv) {
     suffix_init();
     slabs_init(settings.maxbytes, settings.factor, preallocate);
 
-    /* managed instance? alloc and zero a bucket array */
-    if (settings.managed) {
-        buckets = malloc(sizeof(int) * MAX_BUCKETS);
-        if (buckets == 0) {
-            fprintf(stderr, "failed to allocate the bucket array");
-            exit(EXIT_FAILURE);
-        }
-        memset(buckets, 0, sizeof(int) * MAX_BUCKETS);
-    }
-
     /*
      * ignore SIGPIPE signals; we can use errno==EPIPE if we
      * need that information
@@ -3163,18 +3052,22 @@ int main (int argc, char **argv) {
 
     /* create unix mode sockets after dropping privileges */
     if (settings.socketpath != NULL) {
+        errno = 0;
         if (server_socket_unix(settings.socketpath,settings.access)) {
-          fprintf(stderr, "failed to listen\n");
+          fprintf(stderr, "failed to listen on UNIX socket: %s\n", settings.socketpath);
+          if (errno != 0)
+              perror("socket listen");
           exit(EXIT_FAILURE);
         }
     }
 
     /* create the listening socket, bind it, and init */
     if (settings.socketpath == NULL) {
-        int udp_port;
-
-        if (server_socket(settings.port, 0)) {
-            fprintf(stderr, "failed to listen\n");
+        errno = 0;
+        if (settings.port && server_socket(settings.port, 0)) {
+            fprintf(stderr, "failed to listen on TCP port %d\n", settings.port);
+            if (errno != 0)
+                perror("tcp listen");
             exit(EXIT_FAILURE);
         }
         /*
@@ -3183,11 +3076,13 @@ int main (int argc, char **argv) {
          * then daemonise if needed, then init libevent (in some cases
          * descriptors created by libevent wouldn't survive forking).
          */
-        udp_port = settings.udpport ? settings.udpport : settings.port;
 
         /* create the UDP listening socket and bind it */
-        if (server_socket(udp_port, 1)) {
+        errno = 0;
+        if (settings.udpport && server_socket(settings.udpport, 1)) {
             fprintf(stderr, "failed to listen on UDP port %d\n", settings.udpport);
+            if (errno != 0)
+                perror("udp listen");
             exit(EXIT_FAILURE);
         }
     }
